#if !defined( MAIN_H )
#define MAIN_H

/*
 * mdp -- A command-line based markdown presentation tool.
 * Copyright (C) 2018 Michael Goehler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "parser.h"
#include "viewer.h"

#define MDP_VER_MAJOR 1
#define MDP_VER_MINOR 0
#define MDP_VER_REVISION 15

#endif // !defined( MAIN_H )
